// SNF4CGP/JobPool.hpp
// Copyright (C) 2009 ARM Research Labs, LLC
// See www.armresearch.com for more information.
//
// This module maintains a pool of SNF4CGP Jobs associated with the scanner pool
// and is aware of the Output module.
//
// Job objects contain the "how to" for processing any command. They also provide
// a handy place for all of the activity to be recoreded during the processing of
// a Job and they preserve the buffer allocation for that in order to minimize
// memory fragmentation and time spent allocating and deallocating memory.
//
// When a Job is requested one is returned from the pool or if none are
// currently available in the pool an additional Job is allocated.
//
// Each Job object is assigned a command. Then (usually) a worker thread will call
// the doIt() method to get the job done.
//
// When the Job is complete the last thing it does is hand itself off to the Output
// module for processing and recycling -- just before it returns from doIt().

#ifndef IncludedJobPool
#define IncludedJobPool

#include "../SNF4CGP/SNFMulti/SNFMulti.hpp"
#include "../SNF4CGP/CodeDweller/threading.hpp"

#include "InputProcessor.hpp"
#include "ScannerPool.hpp"

#include <string>
#include <vector>

using namespace std;

class Job {                                                                     // SNF4CGP Job instance.
  private:
    ScannerPool& Scanners;                                                      // Knows where to get scanners.
    OutputProcessor& Output;
    Command CurrentCommand;                                                     // Has a current comand.
    string OutputBuffer;                                                        // Preserves an output buffer.
    vector<unsigned char> ReadBuffer;                                           // Preserves a file read buffer.
    unsigned int ReadLength;                                                    // How much data in the buffer.

    //// These methods embody how we get jobs done.

    void doWakeUp();
    void doINTF();
    void doFAIL();
    void doFILE();
    void doRead();
    void doScan();
    void doAction();
    void doBypass();
    void doAllow();
    void doReject();
    void doDelete();
    void doHold();

  public:
    Job(ScannerPool& S, OutputProcessor O);                                     // Construct with important links.
    ~Job();                                                                     // Cleanup when destructing.
    void setCommand(Command C);                                                 // Assign a command for this job.
    void doIt();                                                                // Get the job done.
    string& Responses();                                                        // Read the job's report.
    void clear();                                                               // Cleanup for the next command.
};

class JobPool {                                                                 // Pool of SNF4CGP Jobs
  private:
    OutputProcessor* Output_;                                                   // Output processor (for job recycling)
    OutputProcessor& Output();                                                  // Safe access to the Output Processor.
    ScannerPool* Scanners_;                                                     // Scanner Pool.
    ScannerPool& Scanners();                                                    // Safe access to the Scanner Pool.

    Mutex AllocationMutex;                                                      // Syncrhonization point.
    ProductionQueue<Job*> Jobs;                                                 // Recycled jobs holder.
    unsigned int AllocatedJobs;                                                 // Count of jobs to clean up.

    bool Started;                                                               // True if we're ready to go.

  public:
    JobPool();                                                                  // Basic construction.
    ~JobPool();                                                                 // Safe cleanup on destruction.
    init(string Configuration, OutputProcessor& O);                             // Get ready to allocate jobs.
    Job& grab();                                                                // Get a fresh job object.
    void drop(Job& J);                                                          // Drop a used job object.
    stop();                                                                     // Shutdown and clean up.
};

#endif
